import { version } from './package.json';
import { writeFile, mkdir, chmod, readFile } from 'fs/promises';
import { join } from 'path';
import pegjs from 'pegjs';
import esbuild from 'esbuild';
import { transform as babel } from '@babel/core';

const SRC_PATH = join(__dirname, 'src');
const BUILD_PATH = join(SRC_PATH, '__generated__');
const WEB_PATH = join(__dirname, 'docs');
const CLI_BIN = join(__dirname, 'bin', 'bibtex-tidy');

// For BigInt support:
// * No IE version supports BigInt
// * Edge 79+
// * Firefox 68+
// * Safari 14+
// * Chrome 67+
// All tested on browserstack. Bibtext tidy looks and works correctly.

const BROWSER_TARGETS = {
	edge: '79',
	firefox: '68',
	chrome: '67',
	safari: '14',
};

const banner = `/**
 * bibtex-tidy v${version}
 * https://github.com/FlamingTempura/bibtex-tidy
 *
 * DO NOT EDIT THIS FILE. This file is automatically generated
 * using \`npm run build\`. Edit files in './src' then rebuild.
 **/`;

async function generateParser() {
	const peg = await readFile(join(SRC_PATH, 'bibtex.pegjs'), 'utf8');
	const parser = pegjs.generate(peg, { output: 'source' });
	await writeFile(
		join(BUILD_PATH, 'bibtex.pegjs.js'),
		banner + '\nexport default ' + parser
	);
}

async function generateOptionTypes() {
	const { outputFiles } = await esbuild.build({
		entryPoints: [join(SRC_PATH, 'optionDefinitions.ts')],
		write: false,
		format: 'esm',
	});
	const bundle = new TextDecoder().decode(outputFiles[0].contents);
	// Bundle creates an export which eval doesn't know what to do with. Assign to
	// var instead.
	const options = eval(bundle.replace(/^export/m, 'const res = ') + '; res');

	const ts = [];

	ts.push(banner);
	ts.push('export type Options = {');
	for (const opt of options.optionDefinitions) {
		ts.push('\t/**');
		ts.push(`\t * ${opt.title}`);
		if (opt.description) {
			ts.push('\t *');
			for (const line of opt.description) {
				ts.push(`\t * ${line}`);
			}
		}
		ts.push('\t */');
		ts.push(`\t${opt.key}?: ${opt.type};`);
	}
	ts.push('};');
	ts.push('');

	await writeFile(join(BUILD_PATH, 'optionsType.ts'), ts.join('\n'));
}

async function generateVersionFile() {
	await writeFile(
		join(BUILD_PATH, 'version.ts'),
		`export const version = "${version}";`
	);
}

async function buildJSBundle() {
	console.time('JS bundle built');
	const { outputFiles } = esbuild.buildSync({
		entryPoints: ['./src/index.ts'],
		bundle: true,
		write: false,
		format: 'esm',
		banner: { js: banner },
	});
	const bundle = outputFiles[0];
	const result = babel(bundle.text, {
		presets: [['@babel/env', { targets: BROWSER_TARGETS }]],
		compact: false,
	});
	await writeFile(
		'bibtex-tidy.js',
		result.code + `\nmodule.exports = exports.default;`
	);
	console.timeEnd('JS bundle built');
}

async function buildCLI() {
	console.time('CLI built');
	await esbuild.build({
		bundle: true,
		platform: 'node',
		banner: { js: '#!/usr/bin/env node\n' + banner },
		outfile: CLI_BIN,
		target: ['node7'],
		entryPoints: [join(SRC_PATH, 'cli.ts')],
	});
	await chmod(CLI_BIN, 0o755); // rwxr-xr-x
	console.timeEnd('CLI built');
}

async function buildWebBundle() {
	console.time('Web bundle built');
	const { outputFiles } = await esbuild.build({
		platform: 'browser',
		entryPoints: ['./docs/index.ts'],
		bundle: true,
		write: false,
		banner: { js: banner },
	});
	const bundle = outputFiles[0];
	const result = babel(bundle.text, {
		presets: [['@babel/env', { targets: BROWSER_TARGETS }]],
		compact: false,
	});
	await writeFile(join(WEB_PATH, 'bundle.js'), result.code);
	console.timeEnd('Web bundle built');
}

mkdir(BUILD_PATH, { recursive: true })
	.then(() =>
		Promise.all([
			generateParser(),
			generateOptionTypes(),
			generateVersionFile(),
		])
	)
	.then(() => Promise.all([buildJSBundle(), buildCLI(), buildWebBundle()]));
